# Copyright 1999-2014. Parallels IP Holdings GmbH. All Rights Reserved.
# -*- coding: utf-8; mode: python -*-
"""This scripts analyzes current PAM configuration and
perform necessary changes to incorporate Plesk PAM module
into authentication and password settings chains"""

import sys

try: from itertools import chain
except ImportError: from compat.dummy_itertools import chain

try: import logging
except ImportError: import compat.dummy_logging as logging

from traceback import format_exception, format_exception_only

# PAMConfig
from PAMConfig import ConfSet
from PAMConfig.analyzers import AuthConfAnalyzer, PasswordConfAnalyzer
from PAMConfig.modifiers import AuthConfModifier, PasswordConfModifier

from PAMConfig.defaults import \
     KnownPasswordCheckers, KnownNonPasswordCheckers, \
     KnownPasswordSetters, KnownNonPasswordSetters

# common constants and functions
from pam_plesk_common import *

#### Functions which do most of the job :-)
def getAuthChanges(confSet, theModule, checkers, nonCheckers):
    """collects changes which are to be done to add the module
    to auth facility"""

    authCA  = AuthConfAnalyzer(theModule)

    # receive initial list of PAM services detected on this system
    (autonomous, commonDict) = authCA.selectServices(confSet)
    logging.debug("initial services: autonomous = %s, commonDict = %s" % \
                  (autonomous, commonDict))

    # now drop from the list services which already has theModule
    # activated
    autonomous = authCA.dropInstalled(autonomous, confSet, fullPath = True)
    commonTodo = authCA.dropInstalled(commonDict.iterkeys(), confSet, fullPath = True)
    for k in commonDict.keys()[:]: # work on a copy
        if k not in commonTodo:
            del commonDict[k]
    logging.debug("installed dropped: autonomous = %s, commonDict = %s" % \
                  (autonomous, commonDict))

    # now autonomous and commonDict contain only services for which
    # theModule has't been installed yet

    # Now select services for which at least one of known checkers
    # is used
    autonomousCheckers = authCA.selectImportant(autonomous,
                                                confSet, checkers, fullPath = False)
    # autonomousCheckers is a dict with keys being services names
    # and values being positions in the corresponding configuration
    # list where the change is planned

    # drop the selected modules from the list of all
    #
    # __isub__ isn't implemented for our set emulation so be more
    # traditional
    #autonomous -= autonomousCheckers
    autonomous.difference_update(autonomousCheckers.iterkeys())

    logging.debug("autonomous checkers = %s, autonomous = %s" % \
                  (autonomousCheckers,autonomous))

    commonCheckers = authCA.selectImportant(commonDict.iterkeys(),
                                            confSet, checkers, fullPath = False)
    # just as for autonomousCheckers, commonCheckers is a dict
    # with keys being services names and values being corresponding
    # positions

    # move commonCheckers services from commonDict to commonCheckersDict
    commonCheckersDict = {}
    for s in commonCheckers:
        commonCheckersDict[s] = commonDict[s]
        del commonDict[s]

    # at this point autonomousCheckers and commonCheckersDict
    # contains services which _exactly_ need modifications.
    logging.debug("commonCheckersDict = %s, commonDict = %s" % \
                  (commonCheckersDict, commonDict))
    
    # I guess most of the job is done but let's be a little more
    # meticulous

    # drop those which contains only non-checkers
    autonomous = authCA.dropNonImportant(autonomous, confSet,
                                         nonCheckers)
    common = authCA.dropNonImportant(commonDict.iterkeys(), confSet,
                                    nonCheckers)
    # from now on autonomous is also a dict (see autonomousCheckers
    # structure description above). The same is commonTodo
    for k in commonDict.keys()[:]: # work on a copy
        if k not in common:
            del commonDict[k]
    # endfor
    
    logging.debug("autonomous = %s, commonDict = %s" % \
                  (autonomous, commonDict))

    # here what we have at this point:
    # autonomousCheckers and commonCheckers/commonCheckersDict
    # have lists for absolutely necessary changes

    # autonomous and common/commonDict have lists
    # for probable changes

    # now collect the changes list

    changes = []

    for (svc, position) in autonomousCheckers.iteritems():
        style = authCA.getAddStyle(svc, position, confSet, [])
        changes.append((svc, position, style))

    for (svc, position) in commonCheckers.iteritems():
        style = authCA.getAddStyle(svc, position, confSet, commonCheckersDict[svc])
        changes.append((svc, position, style))

    # FIXME: do we really can process uncertain modules exactly as
    # definitive ones?
    for (svc, position) in autonomous.iteritems():
        style = authCA.getAddStyle(svc, position, confSet, [])
        changes.append((svc, position, style))

    for (svc, position) in common.iteritems():
        style = authCA.getAddStyle(svc, position, confSet, commonDict[svc])
        changes.append((svc, position, style))

    return changes
# enddef getAuthChanges

def applyAuthChanges(changes, confSet, theModule):
    # whooh-hooh... We are collected all required changes for 'auth'
    # facility ...
    changes = sortChanges(changes)
    
    logging.debug("changes = %r" % (changes,))

    modifier = AuthConfModifier(theModule)
    modMethods = {
        'required': modifier.addToRequired,
        'requiredTailed': modifier.addToRequiredTailed,
        'sufficient': modifier.addToSufficient
        }
                  
    for (svc, position, style) in changes:
        try:
            meth = modMethods[style]
        except KeyError:
            raise NotImplementedError("do not know how to handle '%s' change style" % style)
        meth(confSet, svc, position)
    # endfor
# enddef applyAuthChanges

def getPasswordChanges(confSet, theModule, plainPasswordSvc, cryptedPasswordSvc, setters, nonSetters):
    pswdCA = PasswordConfAnalyzer(theModule)

    # Should we go through the all services which could potentially
    # change a password... Well, likely. Beside passwd there could be
    # graphical password changers and other weird things :-)
    
    # receive initial list of PAM services detected on this system
    # for the 'password' facility
    (autonomous, commonDict) = pswdCA.selectServices(confSet)
    logging.debug("initial services: autonomous = %s, commonDict = %s" % \
                  (autonomous, commonDict))

    if not autonomous and not commonDict: # FreeBSD-4 case
        return ([], True, True)

    changes = list()

    # On password change pam_unix.so should use already received password, if available.
    # Maybe we should use KnownPasswordSetters instead of ["pam_unix.so"],
    # but I'm not sure this wouldn't break anything.
    unixPam = pswdCA.selectModules(commonDict.iterkeys(), confSet, ["pam_unix.so"], True)
    for (svc, position) in unixPam.iteritems():
        changes.append((svc, position, "try_first_pass"))

    installed = pswdCA.selectImportant(autonomous, confSet,
                                       [theModule], fullPath = True)
    installedCommon = pswdCA.selectImportant(commonDict.iterkeys(), confSet,
                                             [theModule], fullPath = True)

    plainDetected = plainPasswordSvc in installed
    cryptDetected = cryptedPasswordSvc in installed

    # drop from autonomous and common services those
    # for which the module has been already installed
    autonomous.difference_update(installed.iterkeys())

    for k in installedCommon.iterkeys():
        affected = commonDict[k]
        if not plainDetected:
            plainDetected = (plainPasswordSvc == k or plainPasswordSvc in affected)
        if not cryptDetected:
            cryptDetected = (cryptedPasswordSvc == k or cryptedPasswordSvc in affected)
        del commonDict[k]

    # now select from autonomous and commonDict those
    # services which really need to be changed.
    # unfortunately some br^W ... original software vendors
    # put reduntant stuff into configs so we should remove
    # it from the list
    autonomous = pswdCA.dropNonImportant(autonomous, confSet,
                                         nonSetters, True)
    if not plainDetected:
        plainDetected = plainPasswordSvc in autonomous
    if not cryptDetected:
        cryptDetected = cryptedPasswordSvc in autonomous
        
    common = pswdCA.dropNonImportant(commonDict.iterkeys(), confSet,
                                     nonSetters, True)
    for k in common.iterkeys():
        affected = commonDict[k]
        if not plainDetected:
            plainDetected = (plainPasswordSvc == k or plainPasswordSvc in affected)
        if not cryptDetected:
            cryptDetected = (cryptedPasswordSvc == k or cryptedPasswordSvc in affected)
    
    # currently this simply checks if all selected cmodules are "required"
    for (svc, position) in chain(autonomous.iteritems(), common.iteritems()):
        changes.append((svc, position, pswdCA.getAddStyle(svc, position, confSet)))
    # endfor

    return (changes, not plainDetected, not cryptDetected)
# enddef getPasswordChanges

def fixupFtpd(confSet):
    """Adds account and session stacks (pam_permit'ted)
    if they don't exist yet"""
    svc = confSet.getService("ftpd")
    conf = confSet.getConf("ftpd", "account")
    result = False
    if not len(conf):
        svc.append(confSet.createConfAtom("ftpd", "account", "required",
                                          "pam_permit.so"))
        result = True

    conf = confSet.getConf("ftpd", "session")
    if not len(conf):
        svc.append(confSet.createConfAtom("ftpd", "session", "required",
                                          "pam_permit.so"))
        result = True
    return result
# enddef fixupFtpd

def fixupDes(confSet):
    """Checks password service stack and if pam_pwcheck.so and pam_unix2.so aren't
    configured to use either MD5 or Blowfish passwords (i.e. DES is in use)
    then switch to Blowfish"""
    conf = confSet.getConf("passwd", "password")
    result = False
    
    for (atom, src, listpost) in conf:
        if atom.type == atom.TYPE_CONF and atom.module in ('pam_unix2.so', 'pam_pwcheck.so'):
            if not ('md5' in atom.module_args or 'blowfish' in atom.module_args):
                atom.module_args.append('blowfish')
                atom.orig = None
                result = True
            # endif
        # endif
    # endfor
    return result
# enddef
                
def applyPasswordChanges(changes, confSet, theModule):
    changes = sortChanges(changes)

    modifier = PasswordConfModifier(theModule)
    modMethods = {
        'required': modifier.addToRequired,
        'sufficient': modifier.addToSufficient,
        'try_first_pass': modifier.tryFirstPassArg
        }
                  
    for (svc, position, style) in changes:
        try:
            meth = modMethods[style]
        except KeyError:
            raise NotImplementedError("do not know how to handle '%s' change style" % style)
        meth(confSet, svc, position)
    # endfor
# enddef applyPasswordChanges

def main():
    # be polite to our users and handle all exceptions specifically
    try:
        from pam_plesk_common import make_option
        optadd = [
            make_option("-u", "--fixup-ftpd", dest="fixupFtpd", default=False, action="store_true",
                        help="Make fixups for 'ftpd' service (FreeBSD-4.x)"),

            make_option("--fixup-des", dest="fixupDes", default=False, action="store_true",
                        help="Switch to a blowfish encryption if current encryption is DES (SLES9)")

            ]
        (options, args) = parseArgs(optadd)
        setupLoggers(options)
    except SystemExit:
        # in case we exited after usage()
        raise
    except:
        # Any error at this stage IS a programming error.
        (et, ev, tb) = sys.exc_info()
        trace = 'S1:' + ''.join(format_exception(et, ev, tb))
        logging.debug(trace)
        exitmsg(3, "a programming / runtime error (%s)" % \
                   formatExceptionShort(et, ev) )
    # endtry

    # Load given configuration set into memory
    try:
        logging.debug('Loading PAM configuration from (directory="%s", file="%s")' % \
                      (options.confDir, options.confFile))
        confSet = ConfSet(options.confFile, options.confDir)
    except EnvironmentError:
        # Likely a permission or filesystem consistency error
        (et, ev, tb) = sys.exc_info()
        exitmsg(2, formatExceptionShort(et, ev))
    except NotImplementedError:
        # report the error and attach a whole configuration (now
        # it can be structurally dumped from within confSet.
        # it's a better solution than a zipfile, I guess
        (et, ev, tb) = sys.exc_info()
        trace = 'S2:' + ''.join(format_exception(et, ev, tb))
        # TODO: it's good at this point to store entire set of PAM
        # configuration files
        logging.debug(trace)
        exitmsg(4, "some functionality isn't yet implemented (%s)" % \
                   formatExceptionShort(et, ev) )
    except:
        # likely a programming error. Basically the same as for
        # NotImplementedError
        (et, ev, tb) = sys.exc_info()
        trace = 'S2:' + ''.join(format_exception(et, ev, tb))
        # TODO: it's good at this point to store entire set of PAM
        # configuration files
        logging.debug(trace)
        exitmsg(3, "a programming/runtime error (%s)" % \
                   formatExceptionShort(et, ev) )
    # endtry
    logging.info("""\
PAM configuration loaded from (directory="%s", file="%s"). \
Configuration contains %d file(s)\
""" % (options.confDir, options.confFile, len(confSet.files)))

    # Now we need to calculate changes
    try:
        logging.debug("calculating changes for 'auth' stacks")
        changes = getAuthChanges(confSet, options.pleskModule,
                                 KnownPasswordCheckers, KnownNonPasswordCheckers)
        if changes:
            logging.info("calculated %d position(s) to change 'auth' stacks" % len(changes))
            applyAuthChanges(changes, confSet, options.pleskModule)
            logging.info("required 'auth' stack(s) modified (in memory)")


        logging.debug("calculating changes for 'password' stacks")
        (changes, needPlain, needCrypted) = \
                  getPasswordChanges(confSet, options.pleskModule,
                                     PAMPleskPlainService, PAMPleskCryptedService,
                                     KnownPasswordSetters, KnownNonPasswordSetters)
        if changes:
            logging.info("calculated %d position(s) to change 'password' stacks" % len(changes))
        if needPlain:
            svc = confSet.addService(PAMPleskPlainService)
            svc.append(confSet.createConfAtom(PAMPleskPlainService,
                                              "password", "required",
                                              options.pleskModule, ['try_first_pass']))
            logging.info("'password' stack added and configured for service '%s' (in memory)" % \
                         PAMPleskPlainService)
            

        if needCrypted:
            svc = confSet.addService(PAMPleskCryptedService)
            svc.append(confSet.createConfAtom(PAMPleskCryptedService,
                                              "password", "required",
                                              options.pleskModule, ['try_first_pass']))
            logging.info("'password' stack added and configured for service '%s' (in memory)" % \
                         PAMPleskCryptedService)
        if changes:
            applyPasswordChanges(changes, confSet, options.pleskModule)
            logging.info("required 'password' stack(s) modified (in memory)")

        if options.fixupFtpd:
            if fixupFtpd(confSet):
                logging.info("required 'ftpd' service fixups are performed (in memory)")

        if options.fixupDes:
            if fixupDes(confSet):
                logging.info("switch to blowfish encryption for new passwords is performed (in memory)")
                
    except NotImplementedError:
        # report the error and attach a whole configuration (now
        # it can be structurally dumped from within confSet.
        # it's a better solution than a zipfile, I guess
        (et, ev, tb) = sys.exc_info()
        trace = 'S3:' + ''.join(format_exception(et, ev, tb))
        # TODO: it's good at this point to store entire set of PAM
        # configuration files. Now it can be done by dumping
        # confSet
        logging.debug(trace)
        exitmsg(4, "some functionality isn't yet implemented (%s)" % \
                   formatExceptionShort(et, ev) )
    except:
        # likely a programming error. Basically the same as for
        # NotImplementedError
        (et, ev, tb) = sys.exc_info()
        trace = 'S3:' + ''.join(format_exception(et, ev, tb))
        # TODO: it's good at this point to store entire set of PAM
        # configuration files. Now it can be done by dumping
        # confSet
        logging.debug(trace)
        exitmsg(3, "a programming/runtime error (%s)" % \
                   formatExceptionShort(et, ev) )
    # endtry

    # Time to write the changes back to the disk
    try:
        logging.debug("Ready to write changed stacks back to disk")
        confSet.sync()
        logging.info('PAM configuration saved (directory="%s", file="%s")' % \
                     (options.confDir, options.confFile))
    except EnvironmentError:
        # Likely a permission or filesystem consistency error
        (et, ev, tb) = sys.exc_info()
        exitmsg(2, formatExceptionShort(et, ev))
    except NotImplementedError, e:
        # report the error and attach a whole configuration (now
        # it can be structurally dumped from within confSet.
        # it's a better solution than a zipfile, I guess
        (et, ev, tb) = sys.exc_info()
        trace = 'S4:' + ''.join(format_exception(et, ev, tb))
        # TODO: it's good at this point to store entire set of PAM
        # configuration files. Now it can be done by dumping
        # confSet
        logging.debug(trace)
        exitmsg(4, "some functionality isn't yet implemented (%s)" % \
                   formatExceptionShort(et, ev) )
    except:
        # likely a programming error. Basically the same as for
        # NotImplementedError
        (et, ev, tb) = sys.exc_info()
        trace = 'S4:' + ''.join(format_exception(et, ev, tb))
        # TODO: it's good at this point to store entire set of PAM
        # configuration files. Now it can be done by dumping
        # confSet
        logging.debug(trace)
        exitmsg(3, "a programming/runtime error (%s)" % \
                   formatExceptionShort(et, ev) )
    # endtry

    exitmsg(0, "")
# enddef main

if __name__ == '__main__':
    main()
